"""Gateway connecting to ATAG thermostat."""
import logging
import aiohttp

from .const import (RETRIEVE_PATH, UPDATE_PATH, PAIR_PATH, PAIR_REPLY,
                    UPDATE_REPLY, ACC_STATUS)

from .helpers import HostData, HttpConnector, get_data_from_jsonreply
from .errors import ResponseError, RequestError

_LOGGER = logging.getLogger(__name__)
SESSION = 'session'
HOST = 'host'
PORT = 'port'
MAIL = 'mail'
INTERFACE = 'interface'
DEVICE = 'device'


class AtagDataStore:
    """Central data store entity."""

    def __init__(
            self, session=None, host=None, port=None,
            mail=None, interface=None):

        self.initialized = False
        self.paired = False
        self.session = session
        self._connector = None
        # TODO: overlap met HostData eruit slopen // simpeler maken // self.session ?
        self.host_config = {
            HOST: host,
            PORT: port,
            MAIL: mail,
            INTERFACE: interface,
            DEVICE: None
        }
        self.host_data = None
        self.sensordata = {}

    async def async_finalize_init(self):
        """Atag Discovery in case no host provided."""
        if self.host_config[HOST] is None:
            from .discovery import discover_atag
            try:
                _LOGGER.debug("No host provided, attempting discovery...")
                host, device = await discover_atag()
                self.host_config[HOST] = host
                self.host_config[DEVICE] = device
            except RequestError:
                _LOGGER.error("Atag host discovery failed")
                return False
            _LOGGER.debug("Found Atag at %s\nDevice id: %s", host, device)
        if self.host_config[INTERFACE] is None:
            import netifaces
            self.host_config[INTERFACE] = netifaces.gateways()['default'][netifaces.AF_INET][1]
        try:
            self.host_data = HostData(self.host_config)
        except RequestError:
            _LOGGER.error("Initialization failed: Incorrect host data provided!")
        if self.session is None:
            self.session = aiohttp.ClientSession()
        if type(self.session).__name__ == 'ClientSession':
            self._connector = HttpConnector(self.host_data, self.session)
        else:
            _LOGGER.error("Not a valid session: %s", type(self.session).__name__)
        self.initialized = True

    async def async_update(self):
        """Read data from thermostat."""
        if not self.paired:
            _LOGGER.debug("Atag not paired yet - attempting..")
            if not await self.async_check_pair_status():
                _LOGGER.error(
                    "Pairing failed - please confirm pairing on device")
                return
        try:
            json_data = await self._connector.atag_put(
                data=self.host_data.retrieve_msg, path=RETRIEVE_PATH)
            sensordata = get_data_from_jsonreply(json_data)
            if sensordata:
                self.sensordata = sensordata
        except ResponseError as err:
            _LOGGER.warning('Atag sensor failed to update:\n%s', err)

    async def async_set_atag(self, _target_mode=None, _target_temp=None):
        """set mode and/or temperature."""
        json_payload = self.host_data.get_update_msg(
            _target_mode, _target_temp)
        _LOGGER.debug("Updating: Mode:[%s], Temp:[%s]", _target_mode, _target_temp)
        try:
            json_data = await self._connector.atag_put(data=json_payload, path=UPDATE_PATH)
            _LOGGER.debug("Update reply: %s", json_data)
            if not json_data[UPDATE_REPLY][ACC_STATUS] == 2:
                raise ResponseError(
                    "Invalid update reply received: {}".format(json_data))
        except (ResponseError, KeyError) as err:
            _LOGGER.error("Failed to set Atag: %s", err)
            return False

    async def async_check_pair_status(self):
        """Confirm we are authorized."""

        if self.paired:
            return True
        if not self.initialized:
            await self.async_finalize_init()
            if not self.initialized:
                return False
        try:
            json_data = await self._connector.atag_put(data=self.host_data.pair_msg, path=PAIR_PATH)
            status = json_data[PAIR_REPLY][ACC_STATUS]
        except ResponseError as err:
            _LOGGER.error("Pairing failed\nPairmsg: %s\nError: %s",
                          self.host_data.pair_msg, err)
            return False
        if status == 2:
            self.paired = True
            _LOGGER.debug("AtagDataStore paired")
            return True
        if status == 1:
            _LOGGER.warning("Waiting for pairing confirmation")
        elif status == 3:
            _LOGGER.warning("Waiting for pairing confirmation")
        elif status == 0:
            _LOGGER.warning("No status returned from ATAG One")
        _LOGGER.warning("Atag not paired!\n%s", json_data)
        return False

    async def async_close(self):
        """Close the connection"""
        await self._connector.async_close()



"""Error handling for atag_Thermostat."""


class AtagException(Exception):
    """Base error for atag_Thermostat."""


class RequestError(AtagException):
    """
    Invalid request.
    Unable to fulfill request.
    Raised when host or API cannot be reached.
    """


class ResponseError(AtagException):
    """Invalid response."""


class Response404Error(AtagException):
    """Invalid response."""


class SensorNoLongerAvailable(AtagException):
    """Invalid response."""



"""Helpers for the ATAG connection"""

from datetime import datetime, timezone, timedelta
from numbers import Number
import json

import asyncio
from aiohttp import client_exceptions
from .errors import RequestError, ResponseError
from .const import (REQUEST_INFO, MODES, INT_MODES, BOILER_STATES, BOILER_STATUS, BOILER_CONF,
                    DEFAULT_TIMEOUT, ATTR_OPERATION_MODE, INT_CH_CONTROLS,
                    ATTR_REPORT_TIME, RETRIEVE_REPLY, DETAILS, REPORT, SENSOR_TYPES,
                    REPORT_STRUCTURE_INV, HTTP_HEADER, WEATHER_STATUS, WEATHER_STATES)

MAC = 'mac'
HOSTNAME = 'hostname'
MAIL = 'email'
URL = 'url'


def get_data_from_jsonreply(json_response):
    """Return relevant sensor data from json retrieve reply."""
    result = {}
    try:
        _reply = json_response[RETRIEVE_REPLY]
        _reply[DETAILS] = _reply[REPORT][DETAILS]
        for sensor in SENSOR_TYPES:
            datafield = SENSOR_TYPES[sensor][3]
            location = REPORT_STRUCTURE_INV[datafield]
            if sensor in [BOILER_STATUS, BOILER_CONF, ATTR_OPERATION_MODE,
                          ATTR_REPORT_TIME, WEATHER_STATUS, "ch_status", "ch_control_mode",
                          "dhw_status", "device_status",
                          "connection_status", "date_time"]:
                worker = int(_reply[location][datafield])
                result[sensor] = get_state_from_worker(sensor, worker)
            else:
                try:
                    result[sensor] = float(_reply[location][datafield])
                except ValueError:
                    result[sensor] = _reply[location][datafield]
    except KeyError as err:
        raise ResponseError(err)
    return result


def get_state_from_worker(sensor, worker):
    """
    Returns:\n
    Boiler status based on binary indicator.\n
    Operation mode and Weather status from Atag int.\n
    Report time based on seconds from 2000 (UTC).
    """
    if sensor == BOILER_STATUS:
        return BOILER_STATES[worker & 14]
    if sensor == ATTR_OPERATION_MODE:
        return INT_MODES[worker]
    if sensor == "ch_control_mode":
        return INT_CH_CONTROLS[worker]
    if sensor == WEATHER_STATUS:
        return WEATHER_STATES[worker] # list incl status string and icon
    if sensor in [BOILER_CONF, "ch_status", "dhw_status", "device_status", "connection_status"]:
        return [worker, int_to_binary(worker)] # values TBD..
    if sensor in [ATTR_REPORT_TIME, "date_time"]:
        return datetime(2000, 1, 1, tzinfo=timezone.utc) + timedelta(seconds=worker)
    return False


def int_to_binary(worker):
    """Returns binary representation of int (for certain status/config values)."""
    return '{0:b}'.format(worker)


class HttpConnector:
    """HTTP connector to Bosch thermostat."""

    def __init__(self, hostdata, websession):
        """Init of HTTP connector."""
        self.hostdata = hostdata
        self._websession = websession
        self._request_timeout = DEFAULT_TIMEOUT

    async def atag_put(self, data, path):
        """Make a put request to the API."""

        posturl = '{}{}'.format(self.hostdata.baseurl, path)
        try:
            async with self._websession.put(
                    posturl,
                    json=data,
                    headers=HTTP_HEADER,
                    timeout=self._request_timeout) as req:
                data = await req.text()
                json_result = json.loads(data)
                return json_result
        except (client_exceptions.ClientError,
                client_exceptions.ClientConnectorError, asyncio.TimeoutError) as err:
            raise ResponseError("Error putting data Atag: {}".format(err))
        except json.JSONDecodeError as err:
            raise ResponseError("Unable to decode Json response: {}".format(err))

    def set_timeout(self, timeout=DEFAULT_TIMEOUT):
        """Set timeout for API calls."""
        self._request_timeout = timeout

    async def async_close(self):
        """Close the connection"""
        await self._websession.close()


HOST = 'host'
PORT = 'port'
MAIL = 'mail'
INTERFACE = 'interface'
DEVICE = 'device'


class HostData:
    """Connection info store."""

    def __init__(self, host_config):
        host = host_config[HOST]
        port = host_config[PORT]
        interface = host_config[INTERFACE]
        mail = host_config[MAIL]
        if host is None:
            raise RequestError("Invalid/None host data provided")
        import netifaces
        import socket
        self.email = mail
        self.hostname = socket.gethostname()
        self.baseurl = "http://{}:{}/".format(host, port)
        if interface is None:
            interface = netifaces.gateways()['default'][netifaces.AF_INET][1]
        try:
            self.mac = netifaces.ifaddresses(interface)[
                netifaces.AF_LINK][0]['addr'].upper()
        except ValueError:
            raise RequestError("Incorrect interface selected")
        self.set_pair_msg()
        self.set_retrieve_msg()

    def set_retrieve_msg(self):
        """Get and store the constant retrieve payload."""

        json_payload = {
            "retrieve_message": {
                "seqnr": 1,
                "account_auth": {
                    'user_account': self.email,
                    'mac_address': self.mac
                },
                "info": REQUEST_INFO
            }
        }
        self.retrieve_msg = json_payload

    def set_pair_msg(self):
        """Get and store the constant pairing payload."""

        json_payload = {
            "pair_message": {
                "seqnr": 1,
                "account_auth": {
                    'user_account': self.email,
                    'mac_address': self.mac
                },
                "accounts": {
                    "entries": [
                        {
                            "user_account": self.email,
                            "mac_address": self.mac,
                            "device_name": self.hostname,
                            "account_type": 1
                        }
                    ]
                }
            }
        }
        self.pair_msg = json_payload

    def get_update_msg(self, _target_mode=None, _target_temp=None):
        """Get and return the update payload (mode and temp)."""

        _target_mode_int = None
        if _target_mode is None and _target_temp is None:
            raise RequestError("No update data received")
        if _target_mode is not None:
            if _target_mode in MODES:
                _target_mode_int = MODES[_target_mode]
            else:
                raise RequestError(
                    "Invalid update mode: {}".format(_target_mode))
        elif _target_temp is not None and not isinstance(_target_temp, Number):
            raise RequestError(
                "Not a valid temperature: {}".format(_target_temp))

        json_payload = {
            'update_message': {
                'seqnr': 1,
                'account_auth': {
                    'user_account': self.email,
                    'mac_address': self.mac
                },
                'control': {
                    'ch_mode': _target_mode_int,
                    'ch_mode_temp': _target_temp
                }
            }
        }
        return json_payload



"""Base data to initiate an AtagDataStore."""

host_config = {
    "host": "",  # atag IP
    "port": 10000,
    "mail": "",  # email registered in portal
    "scan_interval": 30,
    "interface": "en0",  # interface on which API runs
    "sensors": [  # sensors to test, see const.py
        "temperature",
        "current_temp"
    ]
}



"""Constants for ATAG API."""

DOMAIN = 'atag'
ATAG_HANDLE = 'atag_data'
DATA_LISTENER = 'atag_listener'
REQUEST_INFO = 1 + 8 + 64  # report, control, and details
SIGNAL_UPDATE_ATAG = 'atag_update'
CONF_INTERFACE = 'interface'

STATE_HEAT = 'heat'
STATE_ECO = 'eco'
STATE_AUTO = 'auto'
STATE_MANUAL = 'manual'

DEFAULT_TIMEOUT = 15
DEFAULT_PORT = 10000
DEFAULT_SCAN_INTERVAL = 120

HTTP_HEADER = {
    'Content-type': 'applicaton/x-www-form-urlencoded;charset=UTF-8',
    'Connection': 'Close',
    'User-Agent': 'Mozilla/5.0 (compatible; AtagOneAPI/x; http://atag.one/)'
}

ATTR_CURRENT_TEMPERATURE = 'current_temperature'
ATTR_MAX_TEMP = 'max_temp'
ATTR_MIN_TEMP = 'min_temp'
ATTR_OPERATION_LIST = 'operation_list'
ATTR_OPERATION_MODE = 'operation_mode'
ATTR_OPERATION_MODE_INT = 'ch_mode'
ATTR_TEMPERATURE_SET = 'shown_set_temp'
ATTR_TEMPERATURE = 'temperature'
ATTR_REPORT_TIME = 'report_time'
BOILER_STATUS = 'boiler_status'
BOILER_CONF = 'boiler_config'
UPDATE_MODE = 'update_mode'
UPDATE_TEMP = 'update_temp'
PAIR_PATH = 'pair'
UPDATE_PATH = 'update'
RETRIEVE_PATH = 'retrieve'
UPDATE_REPLY = 'update_reply'
RETRIEVE_REPLY = 'retrieve_reply'
PAIR_REPLY = 'pair_reply'
STATUS = 'status'
REPORT = 'report'
CONTROL = 'control'
DETAILS = 'details'
ACC_STATUS = 'acc_status'
WEATHER_STATUS = 'weather_status'

DEFAULT_NAME = 'Atag One Thermostat'
DEFAULT_MIN_TEMP = 12
DEFAULT_MAX_TEMP = 21

SENSOR_PREFIX = 'Atag '
REPORT_STRUCTURE = {
    STATUS: ["device_id", "device_status", "connection_status", "date_time"],
    REPORT: ["report_time", "burning_hours", "device_errors", "boiler_errors", "room_temp",
             "outside_temp", "dbg_outside_temp", "pcb_temp", "ch_setpoint", "dhw_water_temp",
             "ch_water_temp", "dhw_water_pres", "ch_water_pres", "ch_return_temp",
             "boiler_status", "boiler_config", "ch_time_to_temp", "shown_set_temp",
             "power_cons", "tout_avg", "rssi", "current", "voltage", "charge_status",
             "lmuc_burner_starts", "dhw_flow_rate", "resets", "memory_allocation"],
    DETAILS: ["boiler_temp", "boiler_return_temp", "min_mod_level", "rel_mod_level",
              "boiler_capacity", "target_temp", "overshoot", "max_boiler_temp", "alpha_used",
              "regulation_state", "ch_m_dot_c", "c_house", "r_rad", "r_env", "alpha", "alpha_max",
              "delay", "mu", "threshold_offs", "wd_k_factor ", "wd_exponent",
              "lmuc_burner_hours", "lmuc_dhw_hours", "KP", "KI"],
    CONTROL: ["ch_status", "ch_control_mode", "ch_mode", "ch_mode_duration", "ch_mode_temp",
              "dhw_temp_setp", "dhw_status", "dhw_mode", "dhw_mode_temp", "weather_temp",
              "weather_status", "vacation_duration", "extend_duration", "fireplace_duration"]
}
SENSOR_TYPES = {
    'device_id': ['Atag One ID', '', 'mdi:account-card-details-outline', 'device_id'],
    'device_status': ['Atag One status', '', 'mdi:account-card-details-outline', 'device_status'],
    'connection_status': ['Atag One ID', '', 'mdi:wifi', 'connection_status'],
    'date_time': ['Atag One ID', '', 'mdi:calendar-clock', 'date_time'],
    ATTR_CURRENT_TEMPERATURE: ['Current Temperature', '°C', 'mdi:thermometer', 'room_temp'],
    'outside_temp': ['Outside Temp', '°C', 'mdi:thermometer', 'outside_temp'],
    'outside_temp_avg': ['Average Outside Temperature', '°C', 'mdi:thermometer', 'tout_avg'],
    WEATHER_STATUS: ['Weather Status', '', 'mdi:white-balance-sunny', 'weather_status'],
    'pcb_temp': ['PCB Temperature', '°C', 'mdi:thermometer', 'pcb_temp'],
    ATTR_TEMPERATURE: ['Target Temperature', '°C', 'mdi:thermometer', ATTR_TEMPERATURE_SET],
    ATTR_OPERATION_MODE: ['Operation Mode', '', 'mdi:settings', ATTR_OPERATION_MODE_INT],
    'ch_water_pressure': ['Central Heating Pressure', 'Bar', 'mdi:gauge', 'ch_water_pres'],
    'ch_water_temp': ['CH Water Temperature', '°C', 'mdi:thermometer', 'ch_water_temp'],
    'ch_return_temp': ['CH Return Temperature', '°C', 'mdi:thermometer', 'ch_return_temp'],
    'dhw_water_temp': ['Hot Water Temp', '°C', 'mdi:thermometer', 'dhw_water_temp'],
    'dhw_water_pres': ['Hot Water Pressure', 'Bar', 'mdi:gauge', 'dhw_water_pres'],
    BOILER_STATUS: ['Boiler Status', '', 'mdi:flash', 'boiler_status'],
    BOILER_CONF: ['Boiler Config', '', 'mdi:flash', 'boiler_config'],
    'burning_hours': ['Burning Hours', 'h', 'mdi:fire', 'burning_hours'],
    'voltage': ['Voltage', 'V', 'mdi:flash', 'voltage'],
    'current': ['Current', 'mA', 'mdi:flash-auto', 'current'],
    'flame_level': ['Flame', '%', 'mdi:fire', 'rel_mod_level'],
    ATTR_REPORT_TIME: ['Report Time', '', 'mdi:clock', ATTR_REPORT_TIME],
    "ch_status": ["ch_status", '', 'mdi:flash', "ch_status"],
    "ch_control_mode": ["ch_control_mode", '', 'mdi:flash', "ch_control_mode"],
    "ch_mode_duration": ["ch_mode_duration", '', 'mdi:flash', "ch_mode_duration"],
    "dhw_temp_setp": ["dhw_temp_setpoint", '', 'mdi:flash', "dhw_temp_setp"],
    "dhw_status": ["dhw_status", '', 'mdi:flash', "dhw_status"],
    "dhw_mode": ["dhw_mode", '', 'mdi:flash', "dhw_mode"],
    "dhw_mode_temp": ["dhw_mode_temp", '', 'mdi:flash', "dhw_mode_temp"],
    "weather_temp": ["weather_temp", '', 'mdi:flash', "weather_temp"],
    "weather_status": ["weather_status", '', 'mdi:flash', "weather_status"],
    "vacation_duration": ["vacation_duration", '', 'mdi:flash', "vacation_duration"],
    "extend_duration": ["extend_duration", '', 'mdi:flash', "extend_duration"],
    "fireplace_duration": ["fireplace_duration", '', 'mdi:flash', "fireplace_duration"]
}

REPORT_STRUCTURE_INV = {
    v: i for i in REPORT_STRUCTURE for v in REPORT_STRUCTURE[i]}
BOILER_STATES = {
    14: 'Heating CV & Water',
    12: 'Heating Water',
    10: 'Heating CV',
    8: 'Heating Boiler',
    4: 'Water active',
    2: 'CV active',
    0: 'Idle'
}

CH_CONTROL_MODES = {
    0: 'Thermostat',
    1: 'Weather dependent'
}

# CH_STATUS = {
#     33: 'Heating type: Convectors / Air heating?', # 8
#     1: 'Heating type: Radiators & Floor'
# }

WEATHER_STATES = {
    0: ['Sunny', 'mdi:weather-sunny'],
    1: ['Clear', 'mdi:weather-night'],
    2: ['Rainy', 'mdi:weather-rainy'],
    3: ['Snowy', 'mdi:weather-snowy'],  # Not sure, Atag icons unclear
    4: ['Haily', 'mdi:weather-hail'],  # Not sure, Atag icons unclear
    5: ['Windy', 'mdi:weather-windy'],
    6: ['Misty', 'mdi:weather-fog'],
    7: ['Cloudy', 'mdi:weather-cloudy'],
    8: ['Partly Sunny', 'mdi:weather-partlycloudy'],
    9: ['Partly Cloudy', 'mdi:cloud'],  # Night with clouds..
    10: ['Shower', 'mdi:weather-pouring'],  # Not sure, Atag icons unclear
    11: ['Lightning', 'mdi:weather-lightning'],
    12: ['Hurricane', 'mdi:weather-hurricane'],
    13: ['Unknown', 'mdi:cloud-question']
}

# fix for Google Assistant integration - use HEAT instead of MANUAL
MODES = {STATE_HEAT: 1, STATE_AUTO: 2}
INT_MODES = {v: k for k, v in MODES.items()}

CH_CONTROLS = {'Thermostat': 0, 'Weather based': 1}
INT_CH_CONTROLS = {v: k for k, v in CH_CONTROLS.items()}



"""Automatic discovery of ATAG Thermostat on LAN."""
import asyncio
from .helpers import RequestError
ATAG_UDP_PORT = 11000
LOCALHOST = '0.0.0.0'

class Discovery(asyncio.DatagramProtocol):
    """Discovery class."""
    def __init__(self):
        self.data = asyncio.Future()
    def connection_made(self, transport):
        print("Listening on UDP {}".format(ATAG_UDP_PORT))
    def datagram_received(self, data, addr):
        self.data.set_result([data, addr])


async def discover_atag():
    """Discover Atag on local network."""
    # return format: [b'ONE xxxx-xxxx-xxxx_xx-xx-xxx-xxx (ST)', ('xxx.xxx.x.x', xxxx)]
    trans, proto = await asyncio.get_event_loop().create_datagram_endpoint(
        Discovery,
        local_addr=(LOCALHOST, ATAG_UDP_PORT))
    try:
        result = await proto.data
        host_ip = result[1][0]
        device_id = result[0].decode().split()[1]
        trans.close()
    except:
        trans.close()
        raise RequestError('Host discovery failed')
    return host_ip, device_id

# To dry run / test
#asyncio.get_event_loop().run_until_complete(discover_atag())
